/*
 * Copyright (c) 1999 The Java Apache Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software and design ideas developed by the Java 
 *    Apache Project (http://java.apache.org/)."
 *
 * 4. The names "Cocoon", "Cocoon Servlet" and "Java Apache Project" must 
 *    not be used to endorse or promote products derived from this software 
 *    without prior written permission.
 *
 * 5. Products derived from this software may not be called "Cocoon"
 *    nor may "Cocoon" and "Java Apache Project" appear in their names without 
 *    prior written permission of the Java Apache Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software and design ideas developed by the Java 
 *    Apache Project (http://java.apache.org/)."
 *           
 * THIS SOFTWARE IS PROVIDED BY THE JAVA APACHE PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *           
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Java Apache Project. For more information
 * on the Java Apache Project please see <http://java.apache.org/>.
 */
 
package org.apache.cocoon;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.w3c.dom.*;
import org.apache.cocoon.parser.*;
import org.apache.cocoon.processor.*;

/**
 * The Cocoon servlet.
 *
 * This servlet implements an XML/XSL server side publishing framework to
 * separate different knowledge contexts in different processing layers.
 *
 * @author <a href="mailto:stefano@apache.org">Stefano Mazzocchi</a>
 * @version $Revision: 1.1.1.1 $ $Date: 1999/03/09 23:09:28 $
 */
 
public class Cocoon extends HttpServlet {
    
    public static final String NAME = "Cocoon";
    public static final String VERSION = "1.0";
    public static final String PROPERTIES = "org/apache/cocoon/cocoon.properties";
    public static final String BROWSER = "browser.";

    /**
     * This inner class is used to store the mapping between browser names
     * and those signature fragments used for mapping
     */
    class UserAgent {
        public String name;
        public String signature;
        public UserAgent(String name, String signature) {
            this.name = name;
            this.signature = signature;
        }
    }
    
    Parser parser;
    Vector processors;
    Vector browsers;
    Properties properties;

    /**
     * This method initializes the servlet.
     */       
    public void init(ServletConfig config) throws ServletException {
        super.init(config);

        properties = getProperties(config.getInitParameter("properties"));

        // Create the user agent filter
        browsers = getUserAgents(properties);

        // Set the factory properties
        Factory.setProperties(properties);
        
        // Create the parser
        parser = ParserFactory.create();
        
        // Prepare the processor context
        ProcessorContext context = new ProcessorContext();
        context.setObject("parser", parser);
        
        // Create the processor pipeline
        processors = ProcessorFactory.createPipeline(context);
        
        log("Parser: " + parser.toString());
        Enumeration stages = processors.elements();
        for (int i = 0; stages.hasMoreElements(); i++) {
            log("Processor " + i + ": " + ((Processor) stages.nextElement()).toString());
        }
    }

    /**
     * Look for the properties file and returns the default one
     * found in the jar file in none is specified or found.
     */
    private Properties getProperties(String file) throws ServletException {
        
        Properties p = new Properties();
        
        try {
            p.load(new FileInputStream(file));
        } catch (Exception e) {
            try {
                p.clear();
                p.load(this.getClass().getClassLoader().getResourceAsStream(PROPERTIES));
            } catch (Exception ex) {
                throw new ServletException("Error reading internal resource: " + ex + ". Archive is probably broken.");
            }
        }
       
        return p;
    }

    /**
     * This method maps the "user-Agent" request parameter to a browser name
     * following the search rules defined in the servlet properties.
     */ 
    public Vector getUserAgents(Properties properties) {
        
        Vector browsers = new Vector();
        
        for (int i = 0; ; i++) {
            String filter = properties.getProperty(BROWSER + i);
            if (filter == null) break;
            int index = filter.indexOf('=');
            browsers.addElement(new UserAgent(filter.substring(0, index), filter.substring(index + 1)));
        }
        
        return browsers;
    }
    
    /**
     * This method maps the "user-Agent" request parameter to a browser name
     * following the search rules defined in the servlet properties.
     */ 
    public String mapUserAgent(String agent) {
        Enumeration e = browsers.elements();
        while (e.hasMoreElements()) {
            UserAgent u = (UserAgent) e.nextElement();
            int index = agent.indexOf(u.signature);
            if (index > -1) return u.name;
        }
        return "unknown";
    }
    
    /**
     * This method is used to abstract the resource from the file system like
     * mandated by the Servlet API 2.1 but to let this servlet remain
     * compatible with 2.0 system until 2.1 systems are widely available.
     */ 
    public URLConnection getResourceConnection(String path) {
        try {
            return (new URL("file:" + path)).openConnection();
        } catch (Exception e) {
            return null;
        }
    }        
        
    /**
     * This method is called by the servlet engine to handle the request.
     */ 
    public void service(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {

        long time = System.currentTimeMillis();
        
        // set content type and other response header fields first
        // FIXME: is Cocoon going to return HTML alone? Should we buffer
        // the servlet output stream to know which type it is at the end? (SM)
        response.setContentType("text/html");

        // get output stream
        PrintWriter out = new PrintWriter(response.getWriter());

        try {
            // get requested resource as URLConnection
            URLConnection resource = getResourceConnection(request.getPathTranslated());

            if (resource != null) {
                // create the document by parsing the file
                Document document = parser.parse(new InputStreamReader(resource.getInputStream()));

                // process the document thru the processor pipeline.
                String file = resource.getURL().getFile();
                String path = file.substring(0, file.lastIndexOf('/') + 1);
                ProcessorParameters parameters = new ProcessorParameters();
                parameters.setParameter("path", path);
                parameters.setParameter("browser", mapUserAgent(request.getHeader("user-Agent")));

                Enumeration stages = processors.elements();
                while (stages.hasMoreElements()) {
                    document = ((Processor) stages.nextElement()).process(document, parameters);
                }
                
                // write and format the document tree onto the output stream
                parser.print(document, out);                
            } else {
                error("Could not access the given URI <b>" + request.getRequestURI()
                    + "</b><p><b>NOTE</b>: this servlet cannot work if the requested "
                    + "file is not accessible by the servlet engine. "
                    + "Please, make sure the requested file is readable by "
                    + "this servlet and reachable by the executing system.</p>", out);
            }
        } catch (Exception e) {
            StringWriter buffer = new StringWriter();
            e.printStackTrace(new PrintWriter(buffer));
            error (buffer.toString(), out);
        }
        
        // Evaluate elapsed time
        time = System.currentTimeMillis() - time;
        
        // Print time comment
        out.println("\n<!-- This page was dynamically created by " + getServletInfo() + " in " + time + " milliseconds -->");
        
        // close output stream
        out.close();
        
    }
    
    private void error(String message, PrintWriter out) throws IOException {
        out.print("<HTML><BODY><CENTER>"
            + "<H2>Servlet Error</H2>"
            + "<P><I>" + message + "</I></P>" 
            + "<HR width=\"50%\">Automatically generated by:<BR>"
            + "<B>" + getServletInfo() + "</B>"
            + "</CENTER></BODY></HTML>");
    }
    
    public String getServletInfo() {
        return NAME + " " + VERSION;
    }
}